import json
import os
import pathlib
from typing import Dict
from unittest import TestCase
from unittest.mock import Mock, patch, call
from uuid import uuid4

import requests
from box import Box
from requests import HTTPError, Response

from pycarlo.common.files import BytesFileReader, JsonFileReader
from pycarlo.features.dbt import DbtImporter
from pycarlo.features.dbt.queries import (
    GET_DBT_UPLOAD_URL,
    IMPORT_DBT_MANIFEST,
    IMPORT_DBT_RUN_RESULTS,
    SEND_DBT_ARTIFACTS_EVENT,
    UPLOAD_DBT_MANIFEST,
    UPLOAD_DBT_RUN_RESULTS,
)
from pycarlo.features.pii import PiiFilterer
from pycarlo.features.user import Resource


class DbtImportServiceTest(TestCase):

    manifest_path = f'{pathlib.Path(__file__).parent}/sample_manifest.json'
    run_results_path = f'{pathlib.Path(__file__).parent}/sample_run_results.json'
    logs_path =f'{pathlib.Path(__file__).parent}/sample_logs.txt'

    def setUp(self) -> None:
        self._mock_pii_service = Mock()
        self._mock_pii_service.get_pii_filters_config.return_value = None

    @patch('pycarlo.features.dbt.dbt_importer.http')
    def test_import_run(self, mock_http: Mock):
        # given
        resource = Resource(id=uuid4(), name='Snowflake', type='snowflake')
        mock_user_service = Mock()
        mock_user_service.get_resource.return_value = resource

        def mock_client_responses(**kwargs):
            query = kwargs['query']
            if query == GET_DBT_UPLOAD_URL:
                return Box({
                    'get_dbt_upload_url': f"https://{kwargs['variables']['fileName']}"
                })

        mock_client = Mock(side_effect=mock_client_responses)

        importer = DbtImporter(
            mc_client=mock_client,
            user_service=mock_user_service,
            pii_service=self._mock_pii_service,
        )

        # when
        importer.import_run(
            manifest_path=self.manifest_path,
            run_results_path=self.run_results_path,
            logs_path=self.logs_path,
            resource_id=resource.id
        )

        # verify expected call to user service
        mock_user_service.get_resource.assert_called_once_with(resource.id)

        # verify expected calls to upload artifacts to S3
        self.assertEqual(3, mock_http.upload.call_count)
        mock_http.upload.assert_has_calls([
            call(
                method='put',
                url='https://sample_manifest.json',
                content=JsonFileReader(self.manifest_path).read()
            ),
            call(
                method='put',
                url='https://sample_run_results.json',
                content=JsonFileReader(self.run_results_path).read()
            ),
            call(
                method='put',
                url='https://sample_logs.txt',
                content=BytesFileReader(self.logs_path).read()
            )
        ])

        # verify expected MC client calls
        self.assertEqual(4, mock_client.call_count)
        mock_client.assert_has_calls([
            call(
                query=GET_DBT_UPLOAD_URL,
                variables=dict(
                    projectName='default-project',
                    invocationId='3b44f6e7-0a4a-4c81-8859-468b2d15075e',
                    fileName='sample_manifest.json'
                )
            ),
            call(
                query=GET_DBT_UPLOAD_URL,
                variables=dict(
                    projectName='default-project',
                    invocationId='3b44f6e7-0a4a-4c81-8859-468b2d15075e',
                    fileName='sample_run_results.json'
                )
            ),
            call(
                query=GET_DBT_UPLOAD_URL,
                variables=dict(
                    projectName='default-project',
                    invocationId='3b44f6e7-0a4a-4c81-8859-468b2d15075e',
                    fileName='sample_logs.txt'
                )
            ),
            call(
                query=SEND_DBT_ARTIFACTS_EVENT,
                variables=dict(
                    projectName='default-project',
                    jobName='default-job',
                    invocationId='3b44f6e7-0a4a-4c81-8859-468b2d15075e',
                    artifacts=dict(
                        manifest='sample_manifest.json',
                        runResults='sample_run_results.json',
                        logs='sample_logs.txt'
                    ),
                    resourceId=str(resource.id)
                )
            )
        ])

    def test_import_dbt_manifest(self):
        self._client_mock = Mock(return_value=Box({
            'import_dbt_manifest': {
                'response': {
                    'node_ids_imported': [
                        "model.analytics.metric_types",
                        "model.analytics.recent_metrics",
                        "model.analytics.lineage_nodes"
                    ]
                }
            }
        }))

        service = DbtImporter(
            mc_client=self._client_mock,
            pii_service=self._mock_pii_service,
        )

        manifest_file = os.path.join(pathlib.Path(__file__).parent.resolve(), 'sample_manifest.json')
        node_ids_imported = service.import_dbt_manifest(manifest_file, default_resource='snowflake')

        with open(manifest_file, 'r') as f:
            dbt_manifest = Box(json.load(f))

        self._client_mock.assert_called_once_with(
            query=IMPORT_DBT_MANIFEST,
            variables=dict(
                dbtSchemaVersion='https://schemas.getdbt.com/dbt/manifest/v2.json',
                manifestNodesJson=json.dumps(dbt_manifest.nodes.to_dict()),
                projectName=None,
                defaultResource='snowflake'
            )
        )

        self.assertEqual(
            node_ids_imported,
            ['model.analytics.metric_types', 'model.analytics.recent_metrics', 'model.analytics.lineage_nodes']
        )

    def test_import_dbt_manifest_retry(self):
        def create_responses(*args, **kwargs):
            nodes = json.loads(kwargs['variables']['manifestNodesJson'])

            if len(list(nodes.items())) == 1:
                return Box({
                    'import_dbt_manifest': {
                        'response': {
                            'node_ids_imported': [list(nodes.keys())[0]]
                        }
                    }
                })

            response = Response()
            response.status_code = requests.codes.gateway_timeout

            raise HTTPError(response=response)

        self._client_mock = Mock(side_effect=create_responses)

        importer = DbtImporter(
            mc_client=self._client_mock,
            pii_service=self._mock_pii_service,
        )

        manifest_file = os.path.join(pathlib.Path(__file__).parent.resolve(), 'sample_manifest.json')
        node_ids_imported = importer.import_dbt_manifest(manifest_file)

        # client call will timeout, then will send each one-by-one
        # client will be called a total of 4 times
        self.assertEqual(4, self._client_mock.call_count)

        self.assertEqual(
            node_ids_imported,
            ['model.analytics.metric_types', 'model.analytics.recent_metrics', 'model.analytics.lineage_nodes']
        )

    def test_import_dbt_manifest_retry_bail_out(self):
        def create_responses(*args, **kwargs):
            response = Response()
            response.status_code = requests.codes.gateway_timeout

            raise HTTPError(response=response)

        self._client_mock = Mock(side_effect=create_responses)

        importer = DbtImporter(
            mc_client=self._client_mock,
            pii_service=self._mock_pii_service,
        )

        manifest_file = os.path.join(pathlib.Path(__file__).parent.resolve(), 'sample_manifest.json')

        with self.assertRaises(RuntimeError):
            node_ids_imported = importer.import_dbt_manifest(manifest_file)

        # make_request_v2() will always timeout
        # First request will time out
        # Then the next will time out, which has a batch size of 1, at which point it bails out
        self.assertEqual(2, self._client_mock.call_count)

    def test_import_dbt_run_results(self):
        self._client_mock = Mock(return_value=Box({
            'import_dbt_run_results': {
                'response': {
                    'num_results_imported': 4
                }
            }
        }))

        service = DbtImporter(
            mc_client=self._client_mock,
            pii_service=self._mock_pii_service,
        )

        run_results_file = os.path.join(pathlib.Path(__file__).parent.resolve(), 'sample_run_results.json')
        with open(run_results_file, 'r') as f:
            run_results = Box(json.load(f))

        node_ids_imported = service.import_run_results(run_results_file)

        self._client_mock.assert_called_once_with(
            query=IMPORT_DBT_RUN_RESULTS,
            variables=dict(
                dbtSchemaVersion='https://schemas.getdbt.com/dbt/run-results/v2.json',
                runResultsJson=json.dumps(run_results),
                projectName=None,
                runId=None,
                runLogs=None
            )
        )

        self.assertEqual(node_ids_imported, 4)

    def test_upload_dbt_manifest(self):
        # given
        project_name = 'mydb'
        resource_name = 'mywarehouse'
        batch_size = 2

        logs = []

        self._client_mock = Mock()
        service = DbtImporter(
            mc_client=self._client_mock,
            print_func=lambda m: logs.append(m),
            pii_service=self._mock_pii_service,
        )

        manifest_file = os.path.join(pathlib.Path(__file__).parent.resolve(), 'sample_manifest.json')
        with open(manifest_file, 'r') as f:
            dbt_manifest = Box(json.load(f))

        all_nodes = dbt_manifest.nodes.to_dict()
        batch_1_nodes = dict(list(all_nodes.items())[:batch_size])
        batch_2_nodes = dict(list(all_nodes.items())[batch_size:])

        # when
        service.upload_dbt_manifest(
            dbt_manifest=manifest_file,
            project_name=project_name,
            default_resource=resource_name,
            batch_size=batch_size)

        # verify expected calls to MC client
        calls = self._client_mock.call_args_list
        self.assertEqual(2, len(calls))

        self.assertDictEqual(dict(
            query=UPLOAD_DBT_MANIFEST,
            variables=dict(
                invocationId='3b44f6e7-0a4a-4c81-8859-468b2d15075e',
                batch=1,
                dbtSchemaVersion='https://schemas.getdbt.com/dbt/manifest/v2.json',
                manifestNodesJson=json.dumps(batch_1_nodes),
                projectName='mydb',
                defaultResource='mywarehouse'
            )
        ), calls[0][1])

        self.assertDictEqual(dict(
            query=UPLOAD_DBT_MANIFEST,
            variables=dict(
                invocationId='3b44f6e7-0a4a-4c81-8859-468b2d15075e',
                batch=2,
                dbtSchemaVersion='https://schemas.getdbt.com/dbt/manifest/v2.json',
                manifestNodesJson=json.dumps(batch_2_nodes),
                projectName='mydb',
                defaultResource='mywarehouse'
            )
        ), calls[1][1])

        # verify expected logging
        self.assertEqual(3, len(logs))
        self.assertListEqual([
            'Uploading 3 DBT objects to Monte Carlo for processing. Please wait...',
            'Uploaded 2 objects',
            'Uploaded 3 objects'
        ], logs)

    def test_upload_dbt_run_results(self):
        expected_run_results = JsonFileReader(self.run_results_path).read()
        self.perform_test_upload_run_results(expected_run_results=expected_run_results)

    @patch("pycarlo.features.dbt.DbtImporter._init_pii_filterer")
    def test_upload_dbt_run_results_filtered(self, mock_init_filterer: Mock):
        mock_init_filterer.return_value = PiiFilterer(filters_config={
            'active': [
                {
                    'name': 'thread-id',
                    'pattern': r'Thread-\d{2}'
                }
            ]
        }, include_metrics=False)

        expected_run_results = JsonFileReader(self.run_results_path).read()
        for r in expected_run_results['results']:
            r['thread_id'] = '<filtered:thread-id>'

        self.perform_test_upload_run_results(expected_run_results=expected_run_results)

    def perform_test_upload_run_results(self, expected_run_results: Dict):
        self._client_mock = Mock()
        service = DbtImporter(mc_client=self._client_mock, pii_service=self._mock_pii_service)

        run_results = JsonFileReader(self.run_results_path).read()
        service.upload_run_results(run_results)

        self._client_mock.assert_called_once_with(
            query=UPLOAD_DBT_RUN_RESULTS,
            variables=dict(
                dbtSchemaVersion='https://schemas.getdbt.com/dbt/run-results/v2.json',
                runResultsJson=json.dumps(expected_run_results),
                invocationId='3b44f6e7-0a4a-4c81-8859-468b2d15075e',
                projectName=None,
                runId=None,
                runLogs=None
            )
        )


    @patch('pycarlo.features.dbt.dbt_importer.http')
    @patch("pycarlo.features.dbt.DbtImporter._init_pii_filterer")
    def test_import_run_filtered(self, mock_init_filterer: Mock, mock_http: Mock):
        resource = Resource(id=uuid4(), name='Snowflake', type='snowflake')
        mock_user_service = Mock()
        mock_user_service.get_resource.return_value = resource

        mock_init_filterer.return_value = PiiFilterer(filters_config={
            'active': [
                {
                    'name': 'thread-id',
                    'pattern': r'Thread-\d{2}'
                }
            ]
        }, include_metrics=False)

        def mock_client_responses(**kwargs):
            query = kwargs['query']
            if query == GET_DBT_UPLOAD_URL:
                return Box({
                    'get_dbt_upload_url': f"https://{kwargs['variables']['fileName']}"
                })

        mock_client = Mock(side_effect=mock_client_responses)

        importer = DbtImporter(
            mc_client=mock_client,
            user_service=mock_user_service,
            pii_service=self._mock_pii_service,
        )

        # when
        importer.import_run(
            manifest_path=self.manifest_path,
            run_results_path=self.run_results_path,
            logs_path=self.logs_path,
            resource_id=resource.id
        )

        expected_run_results = JsonFileReader(self.run_results_path).read()
        for r in expected_run_results['results']:
            r['thread_id'] = '<filtered:thread-id>'

        # verify call was properly filtered, we know there are thread ids in sample run results
        mock_http.upload.assert_has_calls([
            call(
                method='put',
                url='https://sample_run_results.json',
                content=expected_run_results
            ),
        ])



from dataclasses import dataclass
from typing import List, Optional
from unittest import TestCase

from pycarlo.features.metadata import AllowBlockList, MetadataFilter, MetadataFiltersContainer, FilterEffectType, FilterType


@dataclass
class DatasetEntry:
    project: str
    dataset: str


def _redshift_encoder(column: str, value: str, type: FilterType) -> str:
    if type == FilterType.REGEXP:
        op = '~'
        encoded_value = f"'^({value})$'"
    elif type == FilterType.PREFIX:
        op = ' LIKE '
        encoded_value = f"'{value}%'"
    else:
        op = '='
        encoded_value = f"'{value}'"
    return f'{column}{op}{encoded_value}'


class TestDatasetFiltering(TestCase):
    DATASETS = [
        DatasetEntry('project_1', 'dataset_1'),
        DatasetEntry('project_1', 'dataset_2'),
        DatasetEntry('project_1', 'dataset_3'),
        DatasetEntry('project_2', 'dataset_1'),
        DatasetEntry('project_2', 'dataset_2'),
        DatasetEntry('project_2', 'dataset_3'),
        DatasetEntry('project_2', 'dataset_4'),
    ]
    DATASETS_FOR_REGEXP = [
        DatasetEntry('project_1', 'dataset_1'),
        DatasetEntry('project_1', 'dataset_2'),
        DatasetEntry('project_2', 'dataset_1'),
        DatasetEntry('project_2', 'ds_2'),
        DatasetEntry('prj_3', 'ds_1'),
        DatasetEntry('prj_3', 'ds_2'),
        DatasetEntry('prj_3', 'ds_3'),
    ]

    def test_block_single_dataset(self):
        # excludes only project_1.dataset_2
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_1', dataset='dataset_2', effect=FilterEffectType.BLOCK),
        ])
        self._validate_blocked_datasets(filters, [DatasetEntry(project='project_1', dataset='dataset_2')])

    def test_block_multi_datasets(self):
        # excludes project_1.dataset_3 and project_2.dataset_1
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_1', dataset='dataset_3', effect=FilterEffectType.BLOCK),
            MetadataFilter(project='project_2', dataset='dataset_1', effect=FilterEffectType.BLOCK),
        ])
        limits = self._validate_blocked_datasets(filters, [
            DatasetEntry(project='project_1', dataset='dataset_3'),
            DatasetEntry(project='project_2', dataset='dataset_1'),
        ])
        self.assertFalse(limits.is_whole_project_blocked('project_1'))
        self.assertFalse(limits.is_whole_project_blocked('project_2'))

    def test_block_single_project(self):
        # excludes all datasets in project_1
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_1', effect=FilterEffectType.BLOCK),
        ])
        blocked = list(filter(lambda d: d.project == 'project_1', self.DATASETS))
        limits = self._validate_blocked_datasets(filters, blocked)
        self.assertTrue(limits.is_whole_project_blocked('project_1'))
        self.assertFalse(limits.is_whole_project_blocked('project_2'))

    def test_block_second_project(self):
        # excludes all datasets in project_2
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_2', effect=FilterEffectType.BLOCK),
        ])
        blocked = list(filter(lambda d: d.project == 'project_2', self.DATASETS))
        limits = self._validate_blocked_datasets(filters, blocked)
        self.assertFalse(limits.is_whole_project_blocked('project_1'))
        self.assertTrue(limits.is_whole_project_blocked('project_2'))

    def test_block_single_project_regexp(self):
        # excludes all datasets in prj_3
        filters = AllowBlockList(filters=[
            MetadataFilter(project='prj_.*', type=FilterType.REGEXP, effect=FilterEffectType.BLOCK),
        ])
        blocked = list(filter(lambda d: d.project.startswith('prj_'), self.DATASETS_FOR_REGEXP))
        limits = self._validate_blocked_datasets(filters, blocked)
        self.assertFalse(limits.is_whole_project_blocked('project_1'))
        self.assertFalse(limits.is_whole_project_blocked('project_2'))
        self.assertTrue(limits.is_whole_project_blocked('prj_3'))

    def test_block_datasets_regexp(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='.*', dataset='ds_.*', type=FilterType.REGEXP, effect=FilterEffectType.BLOCK),
        ])
        self._validate_dataset_prefix(filters)

    def test_block_datasets_prefix(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_2', dataset='ds_', type=FilterType.PREFIX, effect=FilterEffectType.BLOCK),
            MetadataFilter(project='prj_3', dataset='ds_', type=FilterType.PREFIX, effect=FilterEffectType.BLOCK),
        ])
        self._validate_dataset_prefix(filters)

    def _validate_dataset_prefix(self, filters: AllowBlockList):
        blocked = list(filter(lambda d: d.project.startswith('ds_'), self.DATASETS_FOR_REGEXP))
        limits = self._validate_blocked_datasets(filters, blocked)

        self.assertFalse(limits.is_whole_project_blocked('project_1'))
        self.assertFalse(limits.is_whole_project_blocked('project_2'))
        self.assertFalse(limits.is_whole_project_blocked('prj_3'))
        self.assertFalse(limits.is_whole_project_blocked('new_project'))

        self.assertTrue(limits.is_dataset_allowed('project_2', 'dataset_1'))
        self.assertFalse(limits.is_dataset_allowed('project_2', 'ds_2'))
        self.assertFalse(limits.is_dataset_allowed('prj_3', 'ds_1'))
        self.assertFalse(limits.is_dataset_allowed('prj_3', 'ds_2'))

    def test_no_filters(self):
        expected_datasets = self.DATASETS.copy()
        filtered_datasets = self._filter_datasets(self.DATASETS, MetadataFiltersContainer())
        self.assertEqual(expected_datasets, filtered_datasets)

    def test_allow_single_dataset(self):
        # excludes all datasets in project_1
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_2', dataset='dataset_4', effect=FilterEffectType.ALLOW)
        ], default_effect=FilterEffectType.BLOCK)
        limits = self._validate_allowed_datasets(filters,
                                                 [DatasetEntry(project='project_2', dataset='dataset_4')]
                                                 )

        self.assertTrue(limits.is_whole_project_blocked('project_1'))
        self.assertFalse(limits.is_whole_project_blocked('project_2'))

    def test_allow_second_project(self):
        # excludes all datasets in project_1
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_2', effect=FilterEffectType.ALLOW),
        ], default_effect=FilterEffectType.BLOCK)
        allowed = list(filter(lambda d: d.project == 'project_2', self.DATASETS))
        limits = self._validate_allowed_datasets(filters, allowed)

        self.assertTrue(limits.is_whole_project_blocked('project_1'))
        self.assertFalse(limits.is_whole_project_blocked('project_2'))

    def test_sql_query_block_single_project(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_2', effect=FilterEffectType.BLOCK),
        ])
        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertTrue(limits.is_metadata_filtered)

        conditions = limits.get_sql_conditions({
            'project': 'database',
            'dataset': 'schema'
        }, _redshift_encoder)
        self.assertEqual("(NOT(database='project_2'))", conditions)

    def test_sql_query_block_project_prefix(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='prj_', type=FilterType.PREFIX, effect=FilterEffectType.BLOCK),
        ])
        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertTrue(limits.is_metadata_filtered)

        conditions = limits.get_sql_conditions({
            'project': 'database',
            'dataset': 'schema'
        }, _redshift_encoder)
        self.assertEqual("(NOT(database LIKE 'prj_%'))", conditions)

    def test_sql_query_block_project_dataset_prefix(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_name',
                           dataset='ds_',
                           type=FilterType.PREFIX,
                           effect=FilterEffectType.BLOCK
                           ),
        ])
        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertTrue(limits.is_metadata_filtered)

        conditions = limits.get_sql_conditions({
            'project': 'database',
            'dataset': 'schema'
        }, _redshift_encoder)
        self.assertEqual("(NOT(database LIKE 'project_name%' AND schema LIKE 'ds_%'))", conditions)

    def test_sql_query_block_single_dataset(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_2', dataset='dataset_1', effect=FilterEffectType.BLOCK),
        ])
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions({
            'project': 'database',
            'dataset': 'schema'
        }, _redshift_encoder)
        self.assertEqual("(NOT(database='project_2' AND schema='dataset_1'))", conditions)

    def test_sql_query_block_datasets_regexp(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_2',
                           dataset='dataset_.+',
                           type=FilterType.REGEXP,
                           effect=FilterEffectType.BLOCK
                           ),
        ])
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions({
            'project': 'database',
            'dataset': 'schema'
        }, _redshift_encoder)
        self.assertEqual("(NOT(database~'^(project_2)$' AND schema~'^(dataset_.+)$'))", conditions)

    def test_sql_query_block_datasets_regexp_with_exception(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_2',
                           dataset='dataset_.+',
                           type=FilterType.REGEXP,
                           effect=FilterEffectType.BLOCK
                           ),
            MetadataFilter(project='project_2',
                           dataset='dataset_10',
                           effect=FilterEffectType.ALLOW
                           ),
        ])
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions({
            'project': 'database',
            'dataset': 'schema'
        }, _redshift_encoder)
        self.assertEqual("(((database='project_2' AND schema='dataset_10')) OR "
                         "(NOT(database~'^(project_2)$' AND schema~'^(dataset_.+)$')))", conditions)

    def test_sql_query_project_block_single_dataset(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_2', dataset='dataset_1', effect=FilterEffectType.BLOCK),
            MetadataFilter(project='project_1', dataset='dataset_2', effect=FilterEffectType.BLOCK),
        ])
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions(
            project='project_2',
            column_mapping={
                'project': 'database',
                'dataset': 'schema'
            },
            encoder=_redshift_encoder
        )
        self.assertEqual("(NOT(database='project_2' AND schema='dataset_1'))", conditions)

    def test_sql_query_project_block_nothing(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_1', dataset='dataset_2', effect=FilterEffectType.BLOCK),
        ])
        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertFalse(limits.is_project_with_datasets_filtered('project_2'))
        conditions = limits.get_sql_conditions(
            project='project_2',
            column_mapping={
                'project': 'database',
                'dataset': 'schema'
            },
            encoder=_redshift_encoder
        )
        self.assertIsNone(conditions)

    def test_sql_query_block_multi_datasets(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_2', dataset='dataset_1', effect=FilterEffectType.BLOCK),
            MetadataFilter(project='project_1', dataset='dataset_2', effect=FilterEffectType.BLOCK),
        ])
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions({
            'project': 'database',
            'dataset': 'schema'
        }, _redshift_encoder)
        self.assertEqual("("
                         "NOT(database='project_2' AND schema='dataset_1') AND "
                         "NOT(database='project_1' AND schema='dataset_2')"
                         ")", conditions)

    def test_sql_query_allow_multi_datasets(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_2', dataset='dataset_1', effect=FilterEffectType.ALLOW),
            MetadataFilter(project='project_1', dataset='dataset_2', effect=FilterEffectType.ALLOW),
        ], default_effect=FilterEffectType.BLOCK)
        limits = MetadataFiltersContainer(metadata_filters=filters)
        conditions = limits.get_sql_conditions({
            'project': 'database',
            'dataset': 'schema'
        }, _redshift_encoder)
        self.assertEqual("("
                         "(database='project_2' AND schema='dataset_1') OR "
                         "(database='project_1' AND schema='dataset_2')"
                         ")", conditions)

    def test_whole_project_blocked(self):
        filters = AllowBlockList(filters=[
            MetadataFilter(project='project_2', dataset='dataset_1', effect=FilterEffectType.BLOCK),
        ], default_effect=FilterEffectType.BLOCK)
        limits = MetadataFiltersContainer(metadata_filters=filters)
        self.assertTrue(limits.is_metadata_blocked)
        self.assertTrue(limits.is_whole_project_blocked('project_2'))

        query = limits.get_sql_conditions({
            'project': 'database',
            'dataset': 'schema'
        }, _redshift_encoder)
        self.assertTrue(query)

    def _validate_allowed_datasets(self,
                                   filters: AllowBlockList,
                                   expected_datasets: List,
                                   datasets: Optional[List] = None
                          ) -> MetadataFiltersContainer:
        limits = MetadataFiltersContainer(metadata_filters=filters)
        filtered_datasets = self._filter_datasets(datasets or self.DATASETS, limits)
        self.assertEqual(expected_datasets, filtered_datasets)
        return limits

    def _validate_blocked_datasets(self,
                                   filters: AllowBlockList,
                                   blocked_datasets: List,
                                   datasets: Optional[List] = None
                                   ) -> MetadataFiltersContainer:
        expected_datasets = list(filter(lambda d: d not in blocked_datasets, (datasets or self.DATASETS)))
        limits = MetadataFiltersContainer(metadata_filters=filters)
        filtered_datasets = self._filter_datasets(datasets or self.DATASETS, limits)
        self.assertEqual(expected_datasets, filtered_datasets)
        return limits

    @staticmethod
    def _filter_datasets(datasets: List[DatasetEntry], limits: MetadataFiltersContainer) -> List[DatasetEntry]:
        return [d for d in datasets if limits.is_dataset_allowed(d.project, d.dataset)]



import json
from pathlib import Path
from unittest import TestCase

from pycarlo.common.files import BytesFileReader, JsonFileReader, TextFileReader, to_path


class FileReaderTests(TestCase):

    test_file_path = f'{Path(__file__).parent}/data.json'

    def test_readers(self):
        as_bytes = BytesFileReader(self.test_file_path).read()
        as_dict = JsonFileReader(self.test_file_path).read()
        as_str = TextFileReader(self.test_file_path).read()

        self.assertEqual(as_dict, json.loads(as_bytes))
        self.assertEqual(as_dict, json.loads(as_str))
        self.assertEqual(as_bytes, as_str.encode('utf-8'))


class FileUtilsTests(TestCase):

    def test_to_path(self):
        path_as_str = 'some/relative/path'
        path = to_path(path_as_str)

        self.assertIsInstance(path, Path)
        self.assertEqual(path_as_str, str(path))
        self.assertIs(to_path(path), path)



from pathlib import Path
from typing import Optional

import boto3
import requests

from pycarlo.core import Client, Query

MC_CLIENT = Client()
S3_CLIENT = boto3.client('s3')


def upload_insights_to_s3(destination_bucket: str, desired_file_extension: Optional[str] = '.csv') -> None:
    """Example function for listing all insights in an account, and uploading any available to S3 as a CSV."""
    list_insights_query = Query()
    list_insights_query.get_insights()
    for insight in MC_CLIENT(list_insights_query).get_insights:
        report_name = str(Path(insight.name).with_suffix(desired_file_extension))

        if insight.available:
            report_url_query = Query()
            report_url_query.get_report_url(insight_name=insight.name, report_name=report_name)
            report_url = MC_CLIENT(report_url_query).get_report_url.url

            print(f'Uploading {report_name} to {destination_bucket}.')
            S3_CLIENT.upload_fileobj(
                Fileobj=requests.get(url=report_url, stream=True).raw,
                Bucket=destination_bucket,
                Key=report_name
            )


if __name__ == "__main__":
    upload_insights_to_s3(destination_bucket='<BUCKET-NAME>')



import os
from typing import Tuple


class MissingEnvVar(Exception):
    def __init__(self, var_name: str):
        super(MissingEnvVar, self).__init__()
        self.var_name = var_name

    def __str__(self) -> str:
        return f'Missing expected env variable of {self.var_name}'


def ensure_token_env_vars() -> Tuple[str, str]:
    """
    Verifies that the MCD_DEFAULT_API_ID and MCD_DEFAULT_API_TOKEN environment variables are defined.

    :return: a tuple of the variable values in the form (MCD_DEFAULT_API_ID, MCD_DEFAULT_API_TOKEN)
    :raises SystemExit if an expected variable is missing or empty.
    """
    api_id = os.environ.get('MCD_DEFAULT_API_ID')
    api_token = os.environ.get('MCD_DEFAULT_API_TOKEN')
    if not api_id:
        raise MissingEnvVar(var_name='MCD_DEFAULT_API_ID')
    if not api_token:
        raise MissingEnvVar(var_name='MCD_DEFAULT_API_TOKEN')
    return api_id, api_token


def ensure_endpoint_env_var() -> str:
    """
    Verifies that the MCD_API_ENDPOINT environment variable is defined.

    :return: the value of the variable MCD_API_ENDPOINT
    :raises SystemExit if the expected variable is missing or empty.
    """
    api_url = os.environ.get('MCD_API_ENDPOINT')
    if not api_url:
        raise MissingEnvVar(var_name='MCD_API_ENDPOINT')
    return api_url



from typing import Optional, Union, Dict

from pycarlo.core import Client
from pycarlo.features.pii import PiiFilteringFailModeType
from pycarlo.features.pii.pii_filterer import PiiActiveFiltersConfig, PiiActiveFilter
from pycarlo.features.pii.queries import GET_PII_PREFERENCES, GET_PII_FILTERS


class PiiService:

    def __init__(self, mc_client: Optional[Client] = None):
        self._mc_client = mc_client or Client()

    def get_pii_filters_config(self) -> Optional[Dict]:
        prefs = self._mc_client(query=GET_PII_PREFERENCES).get_pii_filtering_preferences
        if not prefs.enabled:
            return None

        fail_closed = prefs.fail_mode.upper() == PiiFilteringFailModeType.CLOSE
        pii_filters = self._mc_client(query=GET_PII_FILTERS).get_pii_filters
        if not pii_filters:
            return None

        return PiiActiveFiltersConfig(
            fail_closed=fail_closed,
            active=[
                PiiActiveFilter(
                    name=f.name,
                    pattern=f.pattern
                )
                for f in pii_filters if f.enabled
            ]
        ).to_dict()
